(* ocamlc -custom unix.cma dirrecurse.ml *)
(* Todd Goldfinger 
   Last Change: May 2006

   This is provided under the GPL 2.0

   The recur function recursively reads through a directory struture
   applying action to every file.  The action function takes a list
   containing names of files that have been processed in reverse order
   so that the current file is at the head of the list.  The second
   parameter is any type you choose and must be passed as the last
   parameter to recur.  The first and last lists contain the name of
   files that must be at the start and end of the file list,
   respectively for each directory.  So if the files in a directory
   are read into the following list,

   ["file10.txt"; "file0.txt"; "file1.txt"; "file2.txt"]

   and first = ["file0.txt"] and last = ["file10.txt"], then the list
   will be reordered as 

   ["file0.txt"; "file1.txt"; "file2.txt"; "file10.txt"]

   Basically, it just removes everything in first and last from middle
   and puts middle inbetween them.  (So, the order of first and last
   is preserved.)  But you need to make sure those files actually
   exist, or you'll get an error.

   Example:
   
   (* Start will be the first file processed and end the last. *)
   let action files abstr = printf "%s\n" (hd files);
   recur "." action ["start"] ["end"] abstr

 *)

open Unix;;
open Printf;;
open List;;

(* http://caml.inria.fr/pub/ml-archives/ocaml-beginners/2003/04/fa58b4b5e6abe03f74622070eddd8a13.en.html *)
let recur dir action first last abstr =
  let rec enforce_order lst =
    let not_in_first e = not (List.mem e first)
    and not_in_last e = not (List.mem e last) in
    let middle = List.filter not_in_last (List.filter not_in_first lst) in
    first @ middle @ last
  and read_files lst han = 
    lst := List.append [(readdir han)] !lst;
    read_files lst han
  and frecurse flst pflst = 
    if (length flst) > 0 then
      let head = (hd flst) in
      let pflstnew = head::pflst in
      if head <> "." && head <> ".." then
        begin
          try
            let finfo = stat head in
            if finfo.st_kind = S_DIR then
              begin
                let han = opendir head in
                chdir head;
                drecurse han;
                chdir "..";
                closedir han;
                frecurse (tl flst) pflstnew
              end
            else
              begin
                action pflstnew abstr; 
                frecurse (tl flst) pflstnew
              end
          with Unix_error(ENOENT, err, parm) ->
            printf "Having issues with file %s, skipping...\n" head;
            frecurse (tl flst) pflstnew
        end
      else
        frecurse (tl flst) pflstnew
  and drecurse han =
    let flst = ref [] in
    try
      read_files flst han
    with End_of_file ->
      frecurse (enforce_order !flst) []  in 
  let han = opendir dir in
  drecurse han;
  closedir han;;


